<?php
/**
 * Unit tests for the webconfiguration tool's object api
 *
 * The tests aim to:
 *   - include the code used by the webconfig tool itself
 *   - import objects from a single configuration file
 *   - insert those objects into a database
 *   - manipulate the objects, using the object API
 *   - undo all changes (ie, import again)
 *   - save the objects to a different file
 *   - compare the resulting files to ensure they are identical
 */
$cfgs = array($HOME.'testconfig.cfg', $HOME.'mangleconfig.cfg');
require_once('devtools/phptap.php');

# Basic setup first. If we fail before the functions start,
# PHP will bomb out and we'll crash out.
if (!empty($main_tap))
	$tap = $main_tap->sub_init('Object API tests');
else {
	$tap = new phptap('Object API tests');
}

$tap = $tap->sub_init('Importing');
require_once('include/import.inc.php');
$tap->pass("Including import.inc.php works");
foreach ($cfgs as $cfg)
	$tap->ok(posix_access($cfg, POSIX_R_OK), "$cfg is readable");

# Importing starts here
foreach ($cfgs as $cfg)
	$obj_array = import_objects($cfg);
$tap->ok(is_array($obj_array), "import step 1 returned an array");
$tap->ok(!empty($obj_array), "The array returned is non-empty");

function check_index_entries($when, $ocount, $tap)
{
	global $index;

	$sub = $tap->sub_init("index entries $when recombobulate");
	foreach ($ocount as $otype => $cnt) {
		$name = sprintf("$cnt $otype%s should be indexed", $cnt > 1 ? 's' : '');
		$sub->ok_eq(count($index[$otype]), $cnt, $name);
	}

	return $sub->done() === 0;
}

function check_obj_array($when, $ocount, $tap)
{
	global $obj_array;

	$sub = $tap->sub_init("obj_array $when recombobulate");
	foreach ($ocount as $otype => $cnt) {
		$name = sprintf("$cnt $otype%s should be in obj_array", $cnt > 1 ? 's' : '');
		$sub->ok_eq(count($obj_array[$otype]), $cnt, $name);
	}

	return $sub->done() === 0;
}

function check_db_tables($ocount)
{
	global $tap;
	$sub = $tap->sub_init('DB tables');
	foreach ($ocount as $otype => $cnt) {
		$db_table = $otype === 'FILE' ? 'file_tbl' : $otype;
		$query = "SELECT count(id) FROM $db_table";
		$result = nacoma_db::exec_query($query);
		$sub->ok($result !== false, $query);
		$row = nacoma_db::fetch_row($result);
		$sub->ok_eq($row[0], $cnt, "DB entries should match count(obj_array($otype))");
		unset($result);
	}
	$ssub = $sub->sub_init("Junction tables");
	$jnc_tables = array('host_parents', 'custom_vars', 'timeperiod_exclude');
	foreach (array('host', 'service', 'contact') as $otype)
		$jnc_tables[] = get_junction_table_name($otype, $otype . "group");

	foreach (array('host', 'service', 'hostescalation', 'serviceescalation') as $otype) {
		$jnc_tables[] = get_junction_table_name($otype, 'contacts');
		$jnc_tables[] = get_junction_table_name($otype, 'contactgroup');
	}

	foreach ($jnc_tables as $table) {
		$result = nacoma_db::exec_query("SELECT count(*) FROM $table");
		$ssub->ok($result !== false, $query);
		$row = nacoma_db::fetch_row($result);
		$ssub->ok($row[0] !== '0', "$table should be non-empty");
		unset($result);
	}

	$ssub->done();
	return $sub->done() === 0;
}

$icount = array
	('host' => 57, 'host_template' => 4, 'hostgroup' => 9,
	 'contact' => 8, 'contactgroup' => 3, 'service_template' => 4,
	 'timeperiod' => 8, 'command' => 88, 'FILE' => 2,
	 );
$ocount = array
	('command' => 88, 'contact' => 8, 'contactgroup' => 3, 'host' => 61,
	 'hostgroup' => 9, 'service' => 176,
	 'timeperiod' => 8, 'FILE' => 2, 'hostdependency' => 1,
	 'servicedependency' => 4,'hostescalation' => 2, 'serviceescalation' => 4,
	 );

check_index_entries('pre', $icount, $tap);
check_obj_array('pre', $ocount, $tap);
recombobulate_object_array($obj_array);
$ixcount = array('service' => 173);
$oxcount = array('service' => 177, 'serviceescalation' => 4, 'servicedependency' => 4);
foreach ($ixcount as $k => $v)
	$icount[$k] = $v;
foreach ($oxcount as $k => $v)
	$ocount[$k] = $v;
check_index_entries('post', $icount, $tap);
check_obj_array('post', $ocount, $tap);

# Now we override the thing we've done so far by running
# the import the way the webconfig would do it
$tap->ok(
	import_config('phptap', $cfgs),
	"import_config('phptap', '" . print_r($cfgs, true) . "'"
);
$IMPORTING=false;

check_db_tables($ocount);
$tap = $tap->sub_end();

# tests to make sure integral variables are validated properly
function test_nonzero_integers($sub, $obj, $nz_ints = array('max_check_attempts', 'retry_interval'))
{
	foreach ($nz_ints as $nzint) {
		$tmp = false;
		if (isset($obj->obj[$nzint]))
			$tmp = $obj->obj[$nzint];
		$obj->obj[$nzint] = 0;
		$sub->ok($obj->validate_object() === false, "Object shouldn't validate with non-zero '$nzint'");
		if ($tmp === false)
			unset($obj->obj[$nzint]);
		else
			$obj->obj[$nzint] = $tmp;
		$sub->ok($obj->validate_object(), "Object should validate after restoring '$nzint' to a sane value");
	}
}

# object api checking
function service_tests($sub, $obj)
{
	$orig_obj = $obj->obj;
	$obj->obj['service_description'] = "testing, testing ;";
	if (!$sub->ok(!$obj->validation_error_count(), "No validation errors prior to failed validate"))
		var_dump($obj->validation_error_list);
	$result = $obj->validate_object();
	$sub->ok($result === false, "Service description with semi-colon shouldn't validate");
	$sub->ok(!empty($obj->validation_error_list), "validation errors are stored in validation_error_list");
	$sub->ok(!empty($obj->validation_error_list['service_description']), "service_description should be in validation_error_list");
	unset($obj->obj['service_description']);
	$sub->ok($obj->validate_object() === false, "Object shouldn't validate without service_description set");
	$obj->obj['service_description'] = 'testing_.testing';
	$sub->ok($obj->validate_object(), "underscores should be allowed in service_description");
	$obj->obj['service_description'] = 'testing.testing';
	$sub->ok($obj->validate_object(), "dots should be allowed in service_description");
	$obj->obj = $orig_obj;
	$sub->ok($obj->validate_object(), "object should validate after restoring service_description");

	# test validating without contacts set (should be inherited from host)
	$contacts = $contact_groups = false;
	if (isset($obj->obj['contacts'])) {
		$contacts = $obj->obj['contacts'];
		unset($obj->obj['contacts']);
	}
	if (isset($obj->obj['contact_groups'])) {
		$contact_groups = $obj->obj['contact_groups'];
		unset($obj->obj['contact_groups']);
	}
	$sub->ok($obj->validate_object(), "object should validate without either contacts or contact_groups set");
	$obj->obj = $orig_obj;
	test_nonzero_integers($sub, $obj);

	# test that all imported objects has host(group)?_name set
	$res = nacoma_db::exec_query('SELECT id FROM service WHERE (host_name <= 0 OR host_name IS NULL) AND (hostgroup_name <= 0 OR hostgroup_name IS NULL) AND name IS NULL');
	foreach ($res->fetchAll(PDO::FETCH_NUM) as $row) {
		$sub->fail("Service {$row[0]} has neither host_name, hostgroup_name nor name");
	}
	unset($res);
}

function host_tests($sub, $obj)
{
	test_nonzero_integers($sub, $obj, array('max_check_attempts'));
}

# tests common for all groups
function group_tests($sub, $obj)
{
	$nobj = $obj->copy_object();
	$sub->ok(!empty($obj->obj['members']), "'members' should be non-empty before cloning");
	$sub->ok(!empty($nobj->obj['members']), "'members' should be non-empty after clone");
	$sub->ok_eq($nobj->obj['members'], $obj->obj['members'], "clone_object() clones members");
	$sub->ok_eq($nobj->obj['alias'], $obj->obj['alias'], "clone_object() clones alias");
	$sub->ok(empty($nobj->obj[$obj->type . '_name']), $obj->type . '_name should be empty in new object');
}

# stupid wrappers. All groups are tested the same way
function hostgroup_tests($sub, $obj) { group_tests($sub, $obj); }
function servicegroup_tests($sub, $obj) {
	group_tests($sub, $obj);
	$sub->ok(count($obj->obj['members']) === 2, "servicegroup should have 2 members");
}
function contactgroup_tests($sub, $obj) { group_tests($sub, $obj); }

function dependency_tests($sub, $obj)
{
	$vars = array('host_name', 'dependent_host_name');
	if ($obj->type === 'servicedependency')
		$vars = array('service', 'dependent_service');
	foreach ($vars as $v) {
		$sub->ok(!empty($obj->obj[$v]), "obj->obj[$v] should be non-empty");
	}
}

function hostdependency_tests($sub, $obj)
{
	dependency_tests($sub, $obj);
	$sub->ok_eq($obj->obj['dependent_host_name'], $obj->master_obj_id, "master_obj_id should be same as dependent_host_name");
}

function servicedependency_tests($sub, $obj)
{
	dependency_tests($sub, $obj);
	$sub->ok_eq($obj->obj['dependent_service'], $obj->master_obj_id, "master_obj_id should be same as dependent_service");
}

function timeperiod_tests($sub, $obj)
{
	$sub->ok(!empty($obj->custom_vars), "timeperiod should have custom vars (exceptions)");
	$saved = $obj->obj;
	$obj->obj['monday'] = '05:00-03:00';
	$sub->ok(!$obj->validate_object(), "timeperiod with weird timespec should not validate");
	$obj->obj = $saved;
	$saved = $obj->custom_vars;
	$obj->custom_vars['day 3'] = '05:00-03:00';
	$sub->ok(!$obj->validate_object(), "timeperiod with weird timespec on exception should not validate");
	$obj->custom_vars = $saved;
}

function serviceescalation_tests($sub, $obj)
{
	# test that all imported objects refers to a service set
	$res = nacoma_db::exec_query('select id from serviceescalation where service <= 0');
	foreach ($res->fetchAll(PDO::FETCH_NUM) as $row) {
		$sub->fail("Serviceescalation {$row[0]} is orphaned");
	}
	unset($res);
}

function host_template_tests($sub, $obj)
{
	$obj->expand_object();
	$sub->ok(isset($obj->custom_vars['_random_var']), "Host template's custom var is set");
	$sub->ok_eq("random_value", $obj->custom_vars['_random_var'], "Host template's custom var has correct value");
	$obj->custom_vars['_random_var'] = 'predictable_value';
	$obj->custom_vars['_random_var2'] = 'random_value2';
	$sub->ok($obj->save_object(), 'Saving host template with custom var is OK');
	$nobj = new host_template_nacoma_Model($obj->id);
	$sub->ok(isset($nobj->custom_vars['_random_var']), "Host template's custom var is set after save&load");
	$sub->ok(isset($nobj->custom_vars['_random_var2']), "Host template's custom var is set after save&load");
	$sub->ok_eq("predictable_value", $nobj->custom_vars['_random_var'], "Host template's custom var has correct value after save&load");
	$sub->ok_eq("random_value2", $nobj->custom_vars['_random_var2'], "Host template's custom var has correct value after save&load");
}

function basic_obj_test($obj_type, $id, $tap)
{
	$sub = $tap->sub_init("$obj_type object operations");
	$class_name = $obj_type . "_nacoma_Model";
	$obj = new $class_name($id);
	$tobj = new $class_name("$id");

	$sub->ok(is_object($obj), "Loading $obj_type object");
	$sub->ok(is_object($tobj), "Loading $obj_type object with textual id works");
	$sub->ok_eq($tobj->id, $obj->id, "textual and numerical id's should be equal");
	$sub->ok_eq($tobj, $obj, "textual and numerical id's load equal objects");
	$sub->ok_eq($obj->type, $obj_type, "$obj->type should be $obj_type");
	$sub->ok_eq($obj->id, $id, "$obj->id should be $id");
	$sub->ok(!empty($obj->properties), "$obj_type object has properties");
	$sub->ok(!empty($obj->obj), "object must be non-empty");

	foreach (array('var_index', 'required', 'special', 'default') as $k) {
		$sub->ok(is_array($obj->properties[$k]), "$obj_type properties[$k] should be an array");
		$sub->ok(!empty($obj->properties[$k]), "$obj_type properties[$k] should not be empty");
	}

	$name_var = false;
	if ($obj->is_template)
		$name_var = 'name';
	elseif (isset($obj->obj[$obj->type . '_name']))
		$name_var = $obj->type . '_name';

	if ($name_var) {
		$nobj = new $class_name($obj->obj[$name_var]);
		if (!$sub->ok(is_numeric($nobj->id), "Loading $obj->type objects by name should work")) {
			$sub->diag("name_var: $name_var");
			$sub->diag($nobj->id);
		}
		$sub->ok_eq($nobj->id, $obj->id, "name-loaded and id-loaded id's should be equal");
		$sub->ok_eq($nobj->obj, $obj->obj, "name-loaded and id-loaded obj's should be equal");
		$nobj = false;
	}
	if (!$sub->ok($obj->validate_object() === true, "object should validate properly")) {
		$sub->diag("--- errors ---");
		$sub->diag($obj->validation_error_list);
		$sub->diag("--- the object ---");
		$sub->diag($obj->obj);
	}

	$obj->obj['foo'] = 0;
	$tap->ok($obj->save_object() === false, "Saving an object with invalid variables should fail");
	$obj = new $class_name($obj->id);

	# start variable propagation sub-suite
	$sub = $sub->sub_init("$obj->type variable propagation");
	$ret = $obj->propagate_variables(array('foo' => 0), array($obj->id));
	$sub->ok(is_array($ret), "Propagating an invalid variable should fail");
	$prop_var = array();
	foreach ($obj->properties['var_index'] as $k => $v) {
		if ($v === $obj->master_obj_identifier)
			continue;
		if ($v === 'int') {
			$val = isset($obj->obj[$k]) ? $obj->obj[$k] + 1 : 1;
			$prop_var[$k] = $val;
			$ret = $obj->propagate_variables(array($k => $val), array($obj->id));
			$sub->ok($ret === true, "Propagating an integer variable should work");
			$nobj = new $class_name($obj->id);
			$sub->ok(isset($nobj->obj[$k]) && $nobj->obj[$k] == $val, "The object should have the new integer value");
			$obj->save_object();
		}
		if ($v === 'string') {
			$val = isset($obj->obj[$k]) ? $obj->obj[$k] . "foo" : "foo";
			$prop_var[$k] = $val;
			$ret = $obj->propagate_variables(array($k => $val), array($obj->id));
			$sub->ok($ret === true, "Propagating text variable should work, got: ".var_export($ret, true));
			$nobj = new $class_name($obj->id);
			$sub->ok(isset($nobj->obj[$k]) && $nobj->obj[$k] == $val, "The object should have the new text value");
			$obj->save_object();
		}
		if (count($prop_var) === 2)
			break;
	}
	if (!empty($prop_var)) {
		$ret = $obj->propagate_variables($prop_var, array($obj->id));
		$nobj = new $class_name($obj->id);
		foreach ($prop_var as $k => $v) {
			$sub->ok(isset($nobj->obj[$k]) && $nobj->obj[$k] == $v, "$k (1/2) should propagate for $obj->type objects");
		}
		$obj->save_object();
	}
	$sub = $sub->sub_end(false);
	# end variable propagation sub-suite

	# make sure object is fully collapsed before this test
	$obj->collapse_object();
	$orig_obj = $obj->obj;
	$obj->expand_object();
	$obj->collapse_object();
	$sub->ok_eq($orig_obj, $obj->obj, "expand -> collapse shouldn't change the object");
	$obj->obj = $orig_obj;

	$object_specific_tests = $obj_type . "_tests";
	if (function_exists($object_specific_tests)) {
		$object_specific_tests($sub, $obj);
	}

	$sub->skip_end();
	return $sub->done() === 0;
}

# Basic object operations
$sub = $tap->sub_init('Basic object operations');
$basic_test_objs =
	array('timeperiod' => 5, 'contact' => 1,
		  'contactgroup' => 1, 'hostgroup' => 1, 'servicegroup' => 2,
		  'hostdependency' => 2, 'servicedependency' => 7,
		  'command' => 1,
		  'host' => 3, 'host_template' => 2,
		  'service_template' => 1, 'service' => 320,
		  'serviceescalation' => 5);
foreach ($basic_test_objs as $otype => $cnt)
	basic_obj_test($otype, $cnt, $sub);
$sub->done();
$tap->done(empty($main_tap));
